"""
Description:
		Class: Conf
Usage:
		from AspireTUI.Classes import Conf
		myConf = Conf(filename, LOGFILE=log_filename, bVerbose=True)
		myConf.SectionName.Variable
Notes:
		If a LOG_CONFIG= is passed and the file exists, it will be read.
		If it does not exist, the default/passed settings will be written into LOG_CONFIG
		See: myLog.settings for details.
========================================================
Created on:		2024 Jan. 08
Created by:		Simon Arjuna Erat
License:		MIT
URL:			https://www.github.com/sri-arjuna/ASPIRE

Based on my TUI & SWARM for the BASH shell © 2011
"""
#################################################################################################################
#####                                           Class: Conf                                                 #####
#################################################################################################################

# Main Section
class Conf:
	def __init__(self, 
			filename: str, 
			filename_config=None, 
			encoding="UTF-8", 
			bVerbose=False,  
			bDual=False, 
			LOGFILE=None
			):
		"""
			Handling configuration files with ease!
			Opimized for IDE supporting dynamic dot notations.
			
			Upon creation you can toggle some behaviour of the class.
			bVerbose:	Will show user & developer and 'os' actions.
			LOGFILE:	Writes and shows according to default settings
			LOG_CONFIG:	

			While accessing an instanced class like this:
			conf.settings.filename				# Returns filename
			conf.settings.filename = "NewName"	# Provides the filename used with this config (rename)
			conf.reload(bVerbose=True)			# Overwriting class creation setting and
			 									# 	reloading conf.filename, ignoring changes, informing user
			conf.reload(bSave=True)				# Save changes now, wait 3 seconds, then re-reads conf.filename
			conf.save()							# Saves current configuration as conf.filename
			conf.save("new.cfg")				# Exports current config to "new.cfg"
			conf.ig.							# Will expand the content of the current configuration

			At any given section or item you can (depending on context):
			conf.ig.add_section("New Section Name")
			conf.ig.New_Section_Name.add_key("bSomeBool")
			conf.ig.New_Section_Name.bSomeBool = False
			
		"""
		# Structure data in subsection
		def _SubSettings(self):	# , filename=filename, encoding=encoding, bVerbose=bVerbose, bDual=bDual, LOGFILE=LOGFILE
			self.filename = filename
			self.encoding = encoding
			self.bVerbose = bVerbose
			self.bDual = bDual
			self.LOGFILE = LOGFILE
		# Prepare settings
		self.settings = _SubSettings()
		#self.settings.filename = filename
		#self.settings.encoding = encoding
		#self.settings.bVerbose = bVerbose
		#self.settings.bDual = bDual
		#self.settings.LOGFILE = LOGFILE
		#self.settings.
		# Imports and Variables
		import configparser as _configparser
		from AspireTUI import tui as _tui
		from AspireTUI import _MSG
		from AspireTUI import Classes as _Classes
		from AspireTUI import path as _uf
		import AspireTUI.lists as _Lists
		self._tui = _tui
		self._msg = _MSG
		self._lists = _Lists
		_known_extensions = {"ini", "cfg", "conf"}

		if LOGFILE:
			ret, log = _Classes.Log( LOGFILE , bVerbose=bVerbose, bDual=True)
			
			if _tui.status(ret, f"{self._msg.word_filesystem_file}"):
				# TODO
				pass
				
			if LOGFILE:
				if _uf.file_exists(LOGFILE, bVerbose=bVerbose):
					if LOGFILE.__getattribute__("Name"):
						self._logfile = _Classes.Log(LOGFILE)
		else:
			# Is empty
			self._logfile = LOGFILE
		self.bVerbose = bVerbose
		self._config = _configparser.ConfigParser()
		
	def _if_Verbose(self, msg: str):
		"""
			
		"""
		if self.settings.bVerbose:
			self._tui.status(True, "Is Verbose")
		elif self.settings.LOGFILE:
			self._tui.status(True, "Is a logfile")
			self._logfile.status()
		else:
			self._tui.status(False)
	
	def _read(self):
		# Prepare message
		tmp_msg = f"{self._msg.cl_conf_ui_reading}: {self.settings.filename}"
		# Prefer LOG over verbose
		if self.settings.LOGFILE:
			self._logfile.INFO(tmp_msg)
		elif self.settings.bVerbose:
			# Fallback to 
			self._tui.status(self._lists.StatusEnum.Work.id, tmp_msg)
		# Do it
		if self._config.read(self.filename, encoding=self.encoding):
			ret = True
		else:
			ret = False
		# Job is done, report...?
		# Prefer LOG over verbose... again!
		tmp_msg = f"{self._msg.cl_conf_ui_read}: {self.settings.filename}"
		if self.settings.LOGFILE:
			if ret:
				self._logfile.INFO(tmp_msg)
			else:
				self._logfile.WARNING(tmp_msg)
		elif self.settings.bVerbose:
			# Fallback to 
			self._tui.status(ret, tmp_msg)
		
	def _save(self, filename:str):
		with open(self.filename, 'w', encoding=self.encoding) as _configfile:
			self._config.write(_configfile)

	def _reload(self):
		self._read(self)

	def _add_section(self, section):
		self._config.add_section(section)
		self._save()

	def _rename_section(self, old_name, new_name):
		self._config[new_name] = self._config[old_name]
		del self._config[old_name]
		self._save()

	def _add_key(self, section, key, value):
		self._config.set(section, key, value)
		self._save()

	def _remove_key(self, section, key):
		self._config.remove_option(section, key)
		self._save()

	def _rename_key(self, section, old_key, new_key):
		value = self._config[section].pop(old_key)
		self._config[section][new_key] = value
		self._save()

	def _get_section(self, section):
		return self._config[section]

	def _get_key(self, section, key):
		return self._config.get(section, key)

	def _set_key(self, section, key, value):
		self._config.set(section, key, value)
		self._save()

	def reload(self):
		self._reload()

	def save(self):
		self._save()

	class _Key:
		def __init__(self, section, key):
			self.section = section
			self.key = key
		@property
		def value(self):
			return self.section.conf._get_key(self.section.section, self.key)
		@value.setter
		def value(self, new_value):
			self.section.conf._set_key(self.section.section, self.key, new_value)
		def rename(self, new_name):
			self.section.conf._rename_key(self.section.section, self.key, new_name)
			return self.section.conf.__getattr__(self.section.section).__getattr__(new_name)
		def remove(self):
			self.section.conf._remove_key(self.section.section, self.key)
			
	class _Section:
		def __init__(self, conf, section):
			self.conf = conf
			self.section = section

		def __getattr__(self, key):
			if key not in self.conf._get_section(self.section):
				self.conf._add_key(self.section, key, "")
			return self.conf._Key(self, key)
	
		@property
		def _sections(self):
			return list(self._config.sections())

		def __getattr__(self, section):
			if section not in self._config:
				self._add_section(section)
			return self.conf._Section(self, section)


	



"""
Description:
		Contains classes
Usage:
		bVerbose = True 					# Show tui.status messages for each action to user
		                                    #
		from AspireTUI import Classes
		myLog = Classes.Log(filename_log, bVerbose=bVerbose )
		myConf = Classes.Conf(ini_config_file, bVerbose=bVerbose)
========================================================
Created on:		2024 Jan. 08
Created by:		Simon Arjuna Erat
License:		MIT
URL:			https://www.github.com/sri-arjuna/ASPIRE

Based on my TUI & SWARM for the BASH shell © 2011
"""
#################################################################################################################
#####                                           Simple structure for Classes                                #####
#################################################################################################################
from AspireTUI.Classes._Conf import Conf
from AspireTUI.Classes._Log import Log
from AspireTUI.Classes._Crashlog import CrashLog



"""
	Description:
					Public functions to easy several tasks with Files
	Provides:
					from AspireTUI.Classes import Conf
					ret_bool, ret_str = cu.reg_value_get(file, key, var, bDual=True)
	Usage:
					from AspireTUI import Classes
					log = Classes.Log(log_file)
					log.settings.LogLevel_ShowUser =
	========================================================
	Created on:		2024 Jan. 01
	Created by:		Simon Arjuna Erat
	License:		MIT
	URL:			https://www.github.com/sri-arjuna/ASPIRE

	Based on my TUI & SWARM for the BASH shell © 2011
"""
#
#	Essential imports
#
#import os as _os
#import sys as _sys
#import re as _re
#import string as _string
from AspireTUI import _MSG
from AspireTUI import tui as _tui
from typing import Union as _Union
#################################################################################################################
#####                                           Class: LOG                                                  #####
#################################################################################################################
#
# 	Strings for internal use
#
SEVERITY =[]
_SEVERITY_TRANSLATED = []
from AspireTUI.lists import LOG_LEVEL as LEVEL
from AspireTUI.lists import LOG_SEVERITY as SEVERITY

_str_Title = "title"
_str_Desc = "desc"
_str_bUseDate = "bUseDate"
_str_bUseDateSections = "bUseDateSections"
_str_bUseTime = "bUseTime"
_str_bVerbose = "bVerbose"
_str_LL_ShowUser = "LL_ShowUser"
_str_LL_SaveLog = "LL_SaveLog"
_str_bTranslated = "bTranslated"

# English "pre-hardcoded" - obsolete because of import ? ! -- when/if it works
#_SEVERITY[]
#SEVERITY[0] = "DEBUG"
#SEVERITY[1] = "INFO"
#SEVERITY[2] = "WARNING"
#SEVERITY[3] = "ERROR"
#SEVERITY[4] = "CRITICAL"
#SEVERITY[5] = "FATAL"

# Prepare for class creation, this is required


#_str_ = ""
#
#	
#
class _Settings:
	def __init__(self):
		"""
		Settings of the LOG file
		"""
		# Set default values
		self.filename = ""
		self.description = ""
		self.title = ""
		self.bUseDate = False
		self.bUseDateSections = True
		self.bUseTime = True
		self.bVerbose = False
		self.LogLevel_ShowUser = 3
		self.LogLevel_SaveLog = 4
		self.Log_Format = "%H.%M.%S.%f"
		self.bTranslated = False
		# Get acccording strings:
		from AspireTUI.lists import LOG_LEVEL as _LEVEL
		
		# Prepare get/set functions
		@property
		def Description(self):
			return self.Description
		@Description.setter
		def Description(self, new_val: str):
			msg = _MSG.cl_log_err_must_str
			if not isinstance(new_val, str):
				self.WARNING(msg)
				raise ValueError(msg)
			self.Description = new_val
		
		@property
		def title(self):
			return self.title
		@title.setter
		def title(self, new_val: str):
			msg = _MSG.cl_log_err_must_str
			if not isinstance(new_val, str):
				self.WARNING(msg)
				raise ValueError(msg)
			self.title = new_val

		@property
		def bUseDate(self):
			return self.bUseDate
		@bUseDate.setter
		def bUseDate(self, new_val: bool):
			msg = _MSG.cl_log_err_must_bool
			if not isinstance(new_val, bool):
				self.WARNING(msg)
				raise ValueError(msg)
			self.bUseDate = new_val
		
		@property
		def bUseDateSections(self):
			return self.bUseDateSections
		@bUseDateSections.setter
		def bUseDateSections(self, new_val: bool):
			msg = _MSG.cl_log_err_must_bool
			if not isinstance(new_val, bool):
				self.WARNING(msg)
				raise ValueError(msg)
			self.bUseDateSections = new_val
		
		@property
		def bUseTime(self):
			return self.bUseTime
		@bUseTime.setter
		def bUseTime(self, new_val: bool):
			msg = _MSG.cl_log_err_must_bool
			if not isinstance(new_val, bool):
				self.WARNING(msg)
				raise ValueError(msg)
			self.bUseTime = new_val
		
		@property
		def bVerbose(self):
			return self.bVerbose
		@bVerbose.setter
		def bVerbose(self, new_val: bool):
			msg = _MSG.cl_log_err_must_bool
			if not isinstance(new_val, bool):
				self.WARNING(msg)
				raise ValueError(msg)
			self.bVerbose = new_val

class Log:
	def __init__(self, filename: str, bVerbose=False, LOG_CONFIG=None, LOGFILE=None, ):
		
		""""
		Logfile Management Made Easy 			\n

		If you have a LOG_CONFIG, you do not need to provide a LOGFILE. \n
		
		from AspireTUI.Classes import Log  	\n
		myLog = Log( str_filename ) 			\n
		myLog.settings.Title = "Title in new logfile"			\n
		myLog.settings.LogLevel_ShowUser = myLog.LEVEL.WARNING	\n
		myLog.save()			# Saves a config file of the log settings. \n
		\n
		myLog.DEBUG(f"Only shown if myLog.settings.LogLevel_ShowUser is: myLog.LEVEL.DEBUG")	\n
		\n
		SEVERITY[0] = "DEBUG" 		\n
		SEVERITY[1] = "INFO" 		\n
		SEVERITY[2] = "WARNING" 	\n
		SEVERITY[3] = "ERROR" 		\n
		SEVERITY[4] = "CRITICAL" 	\n
		SEVERITY[5] = "FATAL" 		\n
		"""		
		# Actual initialize class
		self.settings = _Settings()
		self.settings.filename = filename
		#global LEVEL
		#self.LEVEL = LEVEL
		#for lvl in self.LEVEL:
	#		lvl._
			
		
		if self.settings.bTranslated:
			self._SEVERITY_USE = _MSG._SEVERITY_TRANSLATED
		else:
			self._SEVERITY_USE = SEVERITY

		# Prepare internal quick use:
		len_sev_max = 0
		if self.settings.bTranslated:
			self._internal_severity_use = self.SEVERITY_TRANSLATED
		else:
			self._internal_severity_use = SEVERITY
		
		for lvl in LEVEL:
			# Get the right strings:
			sStr = self._internal_severity_use[lvl.value]				
			len_tmp = len(sStr)
			# Save new max value if current is longer
			if len_sev_max < len_tmp:
				len_sev_max = len_tmp
		# Now save longest count to attrtibue
		self._internal_indent_severity = len_sev_max
		
	def _has_header(cls, bDual=False):
		"""
		INTERNAL:
		Check if log file has the according content at the top of the file.
		- Title
		- Description 		WIP <--------
		"""
		# Init
		this = cls.settings.filename
		from AspireTUI.path import exists as _exist
		ret_bool = False
		ret_msg = f"{_MSG.cl_conf_ui_reading}: {this}"
		# Check
		if _exist(this):
			with open(this,"r", encoding="UTF-8") as fn:
				pass
			# File exists and header verified (TODO)
			ret_bool = True
		else:
			# Files not found
			ret_bool = False
		# Return
		if bDual:
			return ret_bool, ret_msg
		else:
			ret_bool

	
	def _print_log(cls, level: int, *args):
		"""
		INTERNAL:
		Check if a file needs to be opened
		"""
		# Simple things
		log_file = cls.settings.filename
		iLen = cls._internal_indent_severity
		
		# Prepare message from *args
		if len(args) == 1:
			message = args[0]
		else:
			msg_str = args[0]
			message = msg_str % args[1:]
		# Actualy check if the header is required
		cls._has_header()
		
		def __print_log_savefile(cls, level:int, message=message, fn=log_file):
			"""
			
			"""
			return
			"""
			prefix = cls.settings
			output = ""
			
			while open(fn, "a"):
				# TODO <---------------------------------------------------------
				pass
			"""

		# Show to user?
		if cls.settings.LogLevel_SaveLog >= level:
			_tui.status(111, message)
			#pass
		else:
			__print_log_savefile(cls)
	# 
	def DEBUG(self, *args):
		""""
		Show / Save message to user / file
		"""
		self._print_log(LEVEL.DEBUG.value, *args)

	def INFO(self, *args):
		""""
		Show / Save message to user / file
		"""
		self._print_log(LEVEL.INFO.value, *args)
	
	def WARNING(self, *args):
		""""
		Show / Save message to user / file
		"""
		self._print_log(LEVEL.WARNING.value, *args)
		
	def ERROR(self, *args):
		""""
		Show / Save message to user / file
		"""
		self._print_log(LEVEL.ERROR.value, *args)
	
	def CRITICAL(self, *args):
		""""
		Show / Save message to user / file
		"""
		self._print_log(LEVEL.CRITICAL.value, *args)
		
	def FATAL(self, *args):
		""""
		Show / Save message to user / file
		"""
		self._print_log(LEVEL.FATAL.value, *args)



"""
	Description:
					Public functions to easy several tasks with strings (or math)
	Provides:
					from AspireTUI import StringUtills as stew
					stew.date() , stew.time()
					
	========================================================
	Created on:		2023 Nov. 09
	Created by:		Simon Arjuna Erat
	License:		MIT
	URL:			https://www.github.com/sri-arjuna/ASPIRE

	Based on my TUI & SWARM for the BASH shell © 2011
"""
#
#	Essential imports
#
#import os as _os
#import sys as _sys
import re as _re
#import string as _string
from datetime import datetime as _datetime
# 
#	Cross Platform & Advanced usage
#
#import platform as _platform
#import msvcrt as _msvcrt
#import shutil as _shutil
#import subprocess as _subprocess

#
#	Prepare data structures
#
#from collections import namedtuple as _namedtuple
#from dataclasses import dataclass as _dataclass
from typing import Union as _Union
#
#	Internal imports
#
from AspireTUI.__core._MESSAGES import current as _MSG
from AspireTUI.__core.ColorAndText import cat
"""
def _ClassDate():
	date = date()
	time = time()
	now = now()
	logtime = logtime()
def _ClassChange():
	esc2hex = esc2hex()
	conf2dict = conf2dict()
	sec2time = sec2time()
def Menu():
	datetime = _ClassDate()
	change = _ClassChange()
"""
#################################################################################################################
#####                                           String Utils (stew)                                         #####
#################################################################################################################
def date(bNames=False) -> str:
	"""
	Returns date like:
	2024.12.30
	"""
	current_date = _datetime.now()
	formatted_date = current_date.strftime("%Y.%m.%d")
	if bNames:
		formatted_date += f" {current_date.strftime('%A, %B')}"
	return formatted_date

def time(bLong=False) -> str:
	"""
	Returns time like:
	15:30
	"""
	current_time = _datetime.now()
	formatted_time = current_time.strftime("%H:%M:%S")
	if bLong:
		formatted_time += f":{current_time.strftime('%S')}"
	return formatted_time

def now(clean=False, sep="/", bLong=False, bNames=False) -> str:
	"""
	Returns date and time like: \n
	2024.12.30 / 15:45
	"""
	if clean:
		return f"{date(bNames)} {time(bLong)}"
	else:
		return f"{date(bNames)} {sep} {time(bLong)}"

def logtime() -> str:
	"""
	Returns a string to be used for safe-file-names.
	"""
	current_time = _datetime.now()
	formatted_time = current_time.strftime("%Y.%m.%d_%H.%M.%S")
	return formatted_time

def close_html_tags(html_string: str, exclude="") -> str:
	"""
	Close open HTML tags in the provided string.
	Exclude closing "<div>" tags.
	"""
	# Match open tags
	open_tags = _re.findall(r'<([^/][^>]*)>', html_string)

	# Check for open tags from the end of the string
	for tag in reversed(open_tags):
		if exclude:
			if not tag.startswith(exclude):
				# Close the tag
				html_string += f'</{tag}>'
		else:
			html_string += f'</{tag}>'
	# No treatment required 
	return html_string

def esc2hex(strESC:str) -> str:
	"""
	Transforms internal color codes to web-hex-colors. \n
	Returns an empty string if it didnt find something to close. \n
	Example: colorA = esc2hex(cat.fg.white)
	"""
	#from AspireTUI.strings import cat
	if strESC == cat.reset:
		# Handle special case for cat.reset
		return close_html_tags(strESC)
	elif _re.match(r'<color=[^>]+>', strESC):
			# Fallback: Close color tags
			return _re.sub(r'<color=[^>]+>', '</color>', strESC)
	return ""

def conf2dict(config_string) -> dict:
	"""
	Splits the content of a conf file, or env/set outputs to a dictionary
	"""
	tmp_dict = {}
	for line in config_string.split('\n'):
		if '=' in line:
			key, value = line.split('=', 1)
			tmp_dict[key] = value
	return tmp_dict

def sec2time(sec:_Union[int, float], format=None) -> str:
	"""
	Transforms given seconds to proper HH:MM:SS format.
	"""
	if isinstance(sec, int):
		# Default, basic seconds
		format_fmt = "%H:%M:%S"
	elif isinstance(sec, float):
		# It is microseconds
		format_fmt = "%H:%M:%S.%f"
	else:
		# Unsupported
		raise ValueError(_MSG.word_error, _MSG.cl_log_err_must_int, _MSG.cl_log_err_must_float)

	# Prefer passed format over default
	if format is None:
		format = format_fmt

	# Return passed seconds as time according to format
	return _datetime.datetime.utcfromtimestamp(sec).strftime(format)

def num2roman(num: int) -> str:
	"""
	Converts an integer to Roman numeral
	"""
	if not 0 < num:
		raise ValueError("Input must be a positive integer")
	
	from AspireTUI import Lists as _lists
	roman_numerals = _lists.roman_num2roman
	result = ''
	for value, numeral in roman_numerals.items():
		count = num // value
		result += numeral * count
		num %= value

	return result

def roman2num(roman: str) -> int:
	"""
	Converts Roman numeral to integer
	"""
	from AspireTUI import Lists as _lists
	roman_numerals = _lists.roman_roman2num

	result = 0
	prev_value = 0

	for numeral in reversed(roman):
		value = roman_numerals[numeral]
		if value < prev_value:
			result -= value
		else:
			result += value
		prev_value = value

	return result

def char2morse(input_data):
	"""
	Converts characters or a multiline string to Morse code.
	"""
	from AspireTUI.lists import morse_code as _MORSE_CODE_DICT
	if isinstance(input_data, str):
		input_data = input_data.upper().splitlines()
	return '\n'.join([' '.join([_MORSE_CODE_DICT[char] if char in _MORSE_CODE_DICT else char for char in line]) for line in input_data])

def morse2char(input_data):
	"""
	Converts Morse code to characters.
	"""
	from AspireTUI.lists import morse_code as _MORSE_CODE_DICT
	if isinstance(input_data, str):
		input_data = input_data.splitlines()
	morse_dict_reverse = {value: key for key, value in _MORSE_CODE_DICT.items()}
	return '\n'.join([''.join([morse_dict_reverse[code] if code in morse_dict_reverse else code for code in line.split()]) for line in input_data])

#
#	"String" related bools
#
def isList(this, bDual=False):
	"""
	Simple bool. \n

	"""
	# Init
	bRet = False
	sList = None
	# Check
	if isinstance(this, list):
		bRet = True
		try:
			sList = list(this)
		except TypeError(list, this):
			sList = _MSG.cl_log_err_must_list
			bRet = False
	# Return
	if bDual:
		return bRet, sList
	else:
		return bRet



"""
	Description:
					Public functions to easy several tasks with strings (or math)
	Provides:
					from AspireTUI import ConfigUtills as cu
					ret_bool, ret_str = cu.reg_value_get(file,key,var)
					
	========================================================
	Created on:		2023 Nov. 09
	Created by:		Simon Arjuna Erat
	License:		MIT
	URL:			https://www.github.com/sri-arjuna/ASPIRE

	Based on my TUI & SWARM for the BASH shell © 2011
"""
#
#	Essential imports
#
#import os as _os
#import sys as _sys
#import re as _re
#import string as _string
from AspireTUI import _MSG
# 
#	Cross Platform & Advanced usage
#
#import platform as _platform
#import msvcrt as _msvcrt
#import shutil as _shutil
#import subprocess as _subprocess
import winreg as _winreg
import configparser as _configparser
#
#	Prepare data structures
#
#from enum import Enum as _Enum
#from enum import EnumMeta as _EnumMeta
#################################################################################################################
#####                                           Get registry value                                          #####
#################################################################################################################
# Registry access map
_HKEY = {
	"CLASSES_ROOT": _winreg.HKEY_CLASSES_ROOT,
	"CURRENT_USER": _winreg.HKEY_CURRENT_USER,
	"LOCAL_MACHINE": _winreg.HKEY_LOCAL_MACHINE,
	"USERS": _winreg.HKEY_USERS,
	"CURRENT_CONFIG": _winreg.HKEY_CURRENT_CONFIG
}
def reg_value_get(hive_key, key_path, value_name):
	"""
	Attempts to read passed arguments and returns the according value.	\n
	Recommended:\n
	 		success, message = reg_value_get("HKEY_CURRENT_USER", "path/in/HKCU/", "VarName")	\n
			tui.status(success, message)
	"""
	try:
		if hive_key in _HKEY:
			hkey = _HKEY[hive_key]
			# Open the specified registry key
			with _winreg.OpenKey(hkey, key_path) as key:
				# Read the value of the specified name
				value, value_type = _winreg.QueryValueEx(key, value_name)
				return True, f"{value}"
		else:
			return False, _MSG.registry_hkey
	except Exception as e:
		return False, f"{_MSG.word_error}: {e}"

def reg_value_set(hive_key, key_path, value_name, value, value_type=_winreg.REG_SZ):
	"""
	Attempts to set passed value and returns a bool.	\n
	Recommended:\n
	 		success, message = reg_value_set("HKEY_CURRENT_USER", "path/in/HKCU/", "VarName", "value")	\n
			tui.status(success, message)
	"""
	try:
		if hive_key in _HKEY:
			hkey = _HKEY[hive_key]
			with _winreg.OpenKey(hkey, key_path, 0, _winreg.KEY_SET_VALUE) as key:
				_winreg.SetValueEx(key, value_name, 0, value_type, value)
			return True, _MSG.registry_success
		else:
			return False,  _MSG.registry_hkey
	except Exception as e:
		return False, f"{_MSG.word_error}: {e}"

def reg_value_list_key(hive_key, key_path):
	"""
	Attempts to list all key inside passed key_path, returns them as list.	\n
	Recommended:\n
	 		success, message = reg_value_list_key("HKEY_CURRENT_USER", "path/in/HKCU/")	\n
			tui.status(success, message)
	"""
	try:
		if hive_key in _HKEY:
			hkey = _HKEY[hive_key]
			with _winreg.OpenKey(hkey, key_path) as key:
				subkeys_count, _, _ = _winreg.QueryInfoKey(key)
				subkeys = [_winreg.EnumKey(key, i) for i in range(subkeys_count)]
				return True, subkeys
		else:
			return False,  _MSG.registry_hkey
	except Exception as e:
		return False, f"{_MSG.word_error}: {e}"

def reg_value_list_var(hive_key, key_path):
	"""
	Attempts to list all variable names inside passed key_path, returns them as list.	\n
	Recommended:\n
	 		success, message = reg_value_list_var("HKEY_CURRENT_USER", "path/in/HKCU/")	\n
			tui.status(success, message)
	"""
	try:
		if hive_key in _HKEY:
			hkey = _HKEY[hive_key]
			with _winreg.OpenKey(hkey, key_path) as key:
				values_count, _, _ = _winreg.QueryInfoKey(key)
				values = [_winreg.EnumValue(key, i)[0] for i in range(values_count)]
				return True, values
		else:
			return False,  _MSG.registry_hkey
	except Exception as e:
		return False, f"{_MSG.word_error}: {e}"
#################################################################################################################
#####                                           Config (ini) files                                          #####
#################################################################################################################
def ini_read(c_file, c_key, c_variable):
	"""
	Reads c_file as ini file, looks for c_keys (section) and returns the value of c_variable. \n
	Usage: \n
			value = ini_read("path/to/filename", "section_name", "variable_name")
	"""
	config = _configparser.ConfigParser()
	try:
		config.read(c_file)
		if c_key in config and c_variable in config[c_key]:
			return config[c_key][c_variable]
		return None
	except FileNotFoundError as e:
		return False, f"{_MSG.word_error}:\n{_MSG.file_not_found}: {e}"

def ini_write(c_file, c_key, c_variable, c_value):
	"""
	Reads c_file as ini file, looks for c_keys (section) and returns the value of c_variable. \n
	Usage: \n
			if ini_write("path/to/filename", "section_name", "variable_name", "value"):
				tui.status(True, "Success message")
	"""
	config = _configparser.ConfigParser()
	try:
		config.read(c_file)
		if not config.has_section(c_key):
			config.add_section(c_key)
		config.set(c_key, c_variable, c_value)
		with open(c_file, 'w') as config_file:
			config.write(config_file)
		return True
	except Exception as e:
		return False, f"{_MSG.word_error}: {e}"

def ini_list_keys(c_file):
	"""
	Reads c_file as an ini file and returns a list of all section names.
	Usage:
		keys = ini_list_keys("path/to/filename.ini")
	"""
	config = _configparser.ConfigParser()
	try:
		config.read(c_file)
		return config.sections()
	except FileNotFoundError as e:
		return False, f"{_MSG.word_error}:\n{_MSG.file_not_found}: {e}"

def ini_list_vars(c_file, c_key):
	"""
	Reads c_file as an ini file, looks for c_key (section), and returns a list of all variable names in that section.
	Usage:
		vars = ini_list_vars("path/to/filename.ini", "section_name")
	"""
	config = _configparser.ConfigParser()
	try:
		config.read(c_file)
		if c_key in config:
			return config.options(c_key)
		return None
	except FileNotFoundError as e:
		return False, f"{_MSG.word_error}:\n{_MSG.file_not_found}: {e}"



"""
	Description:
					Handles themes, data and references.
	Provides:
					style = _settings["theme-data"] ; print(theme.border_left)	\n
					list = themes.list()	\n
					themes.set(NAME)		\n
					
	========================================================
	Created on:		2023 Nov. 09
	Created by:		Simon Arjuna Erat
	License:		MIT
	URL:			https://www.github.com/sri-arjuna/ASPIRE

	Based on my TUI & SWARM for the BASH shell © 2011
"""
#from .themes import set_custom_theme, list_available_themes, get_current_theme, load_custom_theme


#import json
#from pathlib import Path

# Load a custom theme from a file
#custom_theme_path = Path.home() / ".config" / "my_custom_theme.json"
#custom_theme = load_custom_theme(custom_theme_path)

# Set the custom theme
#set_custom_theme(custom_theme)

# Now users can include the custom theme in the available themes list
#themes = list_available_themes()
#print(themes)

# Get the current theme, including the custom theme
#current_theme = get_current_theme()
#print(current_theme)

#def load_custom_theme(file_path):
#    try:
#        with open(file_path, 'r') as file:
#            return json.load(file)
#    except FileNotFoundError:
#        print(f"Custom theme file not found at: {file_path}")
#        return None
#    except json.JSONDecodeError:
#        print(f"Error decoding JSON in custom theme file: {file_path}")
#        return None

#def set_custom_theme(custom_theme):
    # Validate custom theme structure if needed
    # ...
#    _available_themes["Custom"] = custom_theme

################################################################################################################
#####                                            Imports                                                   #####
################################################################################################################
#
#	Prepare data structures
#
#from collections import namedtuple as _namedtuple
from dataclasses import dataclass as _dataclass
from enum import Enum as _Enum
#
#	Internals
#
from AspireTUI import _settings_console as _settings
from AspireTUI.strings import cat as _cat
################################################################################################################
#####                                            Theme Structure                                           #####
################################################################################################################
@_dataclass
class _ThemeAttributes:
	border_left: str
	border_right: str
	color_fg: str
	color_bg: str
	prompt_read: str
	prompt_select: str
	bar_empty: str
	bar_half: str
	bar_full: str
	title_left: str
	title_right: str
	title_filler: str
	title_bold: bool
	title_underline: bool
	title_italic: bool
	header_left: str
	header_right: str
	header_filler: str

class _ThemesList(_Enum):
	Default = _ThemeAttributes(
		border_left=" ║ ",
		border_right=" ║ ",
		color_fg="white",
		color_bg="blue",
		prompt_read="\076\076",
		prompt_select="\076\076",
		bar_empty=" ",
		bar_half="─",
		bar_full="═",
		title_left=" ╟─",
		title_right="─╢ ",
		title_filler="─",
		title_bold=True,
		title_underline=False,
		title_italic=True,
		header_left=" ╔═",
		header_right="═╗ ",
		header_filler="═"
	)
	Classic = _ThemeAttributes(
		border_left="# |",
		border_right="| #",
		color_fg="white",
		color_bg="light_blue",
		prompt_read="\076",
		prompt_select="\076",
		bar_empty=" ",
		bar_half="_",
		bar_full="=",
		title_left="# |",
		title_right="| #",
		title_filler=" ",
		title_bold=True,
		title_underline=False,
		title_italic=False,
		header_left="# |",
		header_right="| #",
		header_filler=" "
	)
	Float = _ThemeAttributes(
		border_left="",
		border_right="",
		color_fg="black",
		color_bg="white",
		prompt_read="\076",
		prompt_select="\076",
		bar_empty=" ",
		bar_half="_",
		bar_full="=",
		title_left="",
		title_right="",
		title_filler=" ",
		title_bold=False,
		title_underline=True,
		title_italic=False,
		header_left="",
		header_right="",
		header_filler=" "
	)
	Mono = _ThemeAttributes(
		border_left="| |",
		border_right="| |",
		color_fg="white",
		color_bg="black",
		prompt_read="\076",
		prompt_select="\076",
		bar_empty=" ",
		bar_half="_",
		bar_full="=",
		title_left="| |",
		title_right="| |",
		title_filler=" ",
		title_bold=True,
		title_underline=False,
		title_italic=False,
		header_left="| |",
		header_right="| |",
		header_filler=" "
	)
	Admin = _ThemeAttributes(
		border_left=" ║ ",
		border_right=" ║ ",
		color_fg="white",
		color_bg="red",
		prompt_read="\076\076",
		prompt_select="\076\076",
		bar_empty=" ",
		bar_half="─",
		bar_full="═",
		title_left=" ╠═",
		title_right="═╣ ",
		title_filler="═",
		title_bold=True,
		title_underline=True,
		title_italic=False,
		header_left=" ╔═",
		header_right="═╗ ",
		header_filler="═"
	)
	CyberPunk = _ThemeAttributes(
		border_left=" ║ ",
		border_right=" ║ ",
		color_fg="light_magenta",
		color_bg="light_cyan",
		prompt_read="\076\076",
		prompt_select="\076\076",
		bar_empty=" ",
		bar_half="─",
		bar_full="═",
		title_left=" ╠═",
		title_right="═╣ ",
		title_filler="═",
		title_bold=False,
		title_underline=False,
		title_italic=False,
		header_left=" ╔═",
		header_right="═╗ ",
		header_filler="═"
	)
	# Blocks = https://en.wikipedia.org/wiki/Block_Elements
	Blocks = _ThemeAttributes(
		border_left=" █ ",
		border_right=" █ ",
		color_fg="white",
		color_bg="light_blue",
		prompt_read="\076",
		prompt_select="\076",
		bar_empty="░",
		bar_half="▒",
		bar_full="▓",
		title_left=" █▄",
		title_right="▄█ ",
		title_filler="▀",
		title_bold=True,
		title_underline=False,
		title_italic=False,
		header_left=" █▀",
		header_right="▀█ ",
		header_filler="▀"
	)
# Internal themes
#_available_themes = {theme.name: theme.value for theme in _ThemesList}
#################################################################################################################
#####                                           Theme Stuff                                                 #####
#################################################################################################################
def list() -> list:
	"""
	Returns a list of themese
	"""
	return {theme for theme in _ThemesList.__members__}

def set(newTheme: str):
	"""
	Saves passed string as _settings["theme"] and returns True
	If newTheme is not found in _ThemesList, nothing happens but return False.
	"""
	if newTheme in _ThemesList.__members__:
		_settings["theme"] = newTheme
		return True
	else:
		return False

def get():
	"""
	Retrieves current theme from _settings["theme"] and returns an iterable object (not notation)
	"""
	# Get theme values
	theme_raw = _ThemesList[_settings["theme"]].value

	# Apply console codes to color attribbutes
	theme_use = {}
	for attribute, value in theme_raw.__dict__.items():
		if attribute.startswith('_'):  # Skip internal attributes
			continue
		if 'color_fg' in attribute:
			theme_use[attribute] = getattr(_cat.front, value)
		elif 'color_bg' in attribute:
			theme_use[attribute] = getattr(_cat.back, value)
		else:
			theme_use[attribute] = value
	
	return type('ThemeData', (), theme_use)



"""
Description: \n
	Hello Random Coder. \n
	Here you'll find everything eventhough you're not supposed to. \n
	\n
	This is the "Core Engine" if you will. \n
\n
======================================================== \n
File Created on:		2024 Jan. 11 \n
File Created by:		Simon Arjuna Erat \n
License:				MIT \n
URL:					https://www.github.com/sri-arjuna/ASPIRE \n
Based on my TUI & SWARM for the BASH shell © 2011 \n
"""



"""
Search Evaluate Action, © 2016 by Simon Arjuna Erat
---------------------------------------------------

SEA's Complex Data Structure Class (CDSC)
~~ a python dataclass experiment ~~

Abstract:
First part of the class is an enum which limits the "top level" count.
We then have a 3 dimensional array offering various "knot"s "node"s.

Within this 3 dimensional array, everything is a node! Everything has get/set.
Each node must have 1 parent. (a node can have any node as parent, but only one.)
Each node may have one or more children. (other nodes, making this one a knot - for explaning purposes)
Each node must have a node_label
Each node must have a node_type (1 of 3, --> double array, node based tree structure)
Each node might have a value
Each node might have a value_type (as example: str, int, bool, hex, float)
Each node might have content
Each node might have a message

And when I say "x dimensional array", i mean asociated arrays, well "array_name["My House"]".

For the 'structure' of the data, coding wise:
class.
	reload(str_import_db=None, bSafe=False) 	# Re-reads the file, without safing first, then refresh the tree (root)
	save(str_backup_db=None)					# Safe class.settings.filename or export it
	refresh(bVerbose=self.settings.bVerbose)	# Refresh while giving output to console, remove everything below class.<content of root_label>
class.settings. # (each have get/set property)
	bVerbose: bool
	filename: str
	title: str
	description: str
	logfile: str
	confile: str
	save(str_export_conf_file=None)		# Saves everything of class.settings.* (for example incl: class.settings.data.*)
	reload(str_import_conf_file=None, bSafe=False)
class.settings.data.  # The "enum list" is used for top level entries, and to define filenames for the seperate saving/loading. (also: set/get)
	enum: str, list
	prefix: str
	suffix: str
	ext: str
	root_label: str
class.query.
	list(enum,a,b,c)
	search(enum, str)
class.<root_label>.
	Enum1.
		<content of db>, use node parent/children attributes to structure it accordingly.
	Enum2.
		...
		

	
Example Usage:
say: class.settings.data.global is "global"
class.global.EntryOfSettings-Data-Enum.type = 1 # Being "double asociated array"
class.global.EntryOfSettings-Data-Enum[topic][section] = 2 # Beeing "nodes" (optimized for tree or other structures)
class.global.EntryOfSettings-Data-Enum.topic.section.type = 2 	# Later one... would do the same, type refering to node_type - in both meanings

this_list = class.query.list(enum,a,b,c)
this_string = class.query.search(enum, str)
	

important:
what would be the file size of an empty db?
what would be the memory (ram) footprint of the empty db while loaded / open?
In other words, what about would be the so called "overhead"?

"""
